/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef POINTIMPLSSE_HH
#define POINTIMPLSSE_HH

namespace SharpConstruct
{
	Point3D::Point3D()
	: data_( _mm_setzero_ps() )
	{}
	Point3D::Point3D( const float x, const float y, const float z )
	: data_( _mm_set_ps( 0, z, y, x ) )
	{}
	Point3D::Point3D( const __m128& d )
	: data_( d )
	{}
	
	float& Point3D::X()
	{
		return component_( 0 );
	}
	float& Point3D::Y()
	{
		return component_( 1 );
	}
	float& Point3D::Z()
	{
		return component_( 2 );
	}
	
	float Point3D::X() const
	{
		return component_( 0 );
	}
	float Point3D::Y() const
	{
		return component_( 1 );
	}
	float Point3D::Z() const
	{
		return component_( 2 );
	}

	const __m128& Point3D::Data() const
	{
		return data_;
	}
	
	void Point3D::operator+=( const Point3D& a )
	{
		data_ += a.data_;
	}
	void Point3D::operator-=( const Point3D& a )
	{
		data_ -= a.data_;
	}
	void Point3D::operator*=( const Point3D& a )
	{
		data_ *= a.data_;
	}
	void Point3D::operator/=( const Point3D& a )
	{
		data_ /= a.data_;
	}
	void Point3D::operator+=( const float a )
	{
		data_ += _mm_set1_ps( a );
	}
	void Point3D::operator-=( const float a )
	{
		data_ -= _mm_set1_ps( a );
	}
	void Point3D::operator*=( const float a )
	{
		data_ *= _mm_set1_ps( a );
	}
	void Point3D::operator/=( const float a )
	{
		data_ /= _mm_set1_ps( a );
	}
	
	Point3D Point3D::operator+( const Point3D& a ) const
	{
		return Point3D( data_ + a.data_ );
	}
	Point3D Point3D::operator-( const Point3D& a ) const
	{
		return Point3D( data_ - a.data_ );
	}
	Point3D Point3D::operator*( const Point3D& a ) const
	{
		return Point3D( data_ * a.data_ );
	}
	Point3D Point3D::operator/( const Point3D& a ) const
	{
		return Point3D( data_ / a.data_ );
	}
	Point3D Point3D::operator+( const float a ) const
	{
		return Point3D( data_ + _mm_set1_ps( a ) );
	}
	Point3D Point3D::operator-( const float a ) const
	{
		return Point3D( data_ - _mm_set1_ps( a ) );
	}
	Point3D Point3D::operator*( const float a ) const
	{
		return Point3D( data_ * _mm_set1_ps( a ) );
	}
	Point3D Point3D::operator/( const float a ) const
	{
		return Point3D( data_ / _mm_set1_ps( a ) );
	}

	inline float& Point3D::component_( const unsigned i )
	{
		return ( ( float* )&data_ )[ i ];
	}
	inline float Point3D::component_( const unsigned i ) const
	{
		return ( ( float* )&data_ )[ i ];
	}

	Point3D Cross( const Point3D& a, const Point3D& b )
	{
		return Point3D( a.Y() * b.Z() - a.Z() * b.Y(),
				a.Z() * b.X() - a.X() * b.Z(),
				a.X() * b.Y() - a.Y() * b.X() );
	}

	float Distance( const Point3D& a, const Point3D& b )
	{
		register __m128 p1 = a.Data(), shuff, dis;
		p1 = _mm_sub_ps( p1, b.Data() );
		p1 = _mm_mul_ps( p1, p1 );
		shuff = _mm_shuffle_ps( p1, p1, _MM_SHUFFLE( 1, 2, 3, 0 ) );
		dis = shuff;
		shuff = _mm_shuffle_ps( p1, p1, _MM_SHUFFLE( 2, 3, 0, 1 ) );
		dis = _mm_add_ss( dis, shuff );
		shuff = _mm_shuffle_ps( p1, p1, _MM_SHUFFLE( 3, 0, 1, 2 ) );
		dis = _mm_add_ss( dis, shuff );
		dis = _mm_rsqrt_ss( dis );
		dis = _mm_rcp_ss( dis );
		float d;
		_mm_store_ss( &d, dis );
		return d;
	}

	float Dot( const Point3D& a, const Point3D& b )
	{
		return HorizontalAdd( a * b );
	}

	float HorizontalAdd( const Point3D& a )
	{
		float ret( 0 );
		register __m128 t( a.Data() );
		t = _mm_add_ps( t, _mm_shuffle_ps( t, t, _MM_SHUFFLE( 2, 1, 0, 3 ) ) );
		t = _mm_add_ps( t, _mm_shuffle_ps( t, t, _MM_SHUFFLE( 2, 2, 2, 2 ) ) );
		_mm_store_ss( &ret, t );
		return ret;
	}

	Point3D Midpoint( const Point3D& a, const Point3D& b )
	{
		return ( a + b ) / 2;
	}

	void Normalize( Point3D& a )
	{
		register __m128 tmp( a.Data() );
		register __m128 length;
		register __m128 shuff;
		// Square all the components
		tmp = _mm_mul_ps( tmp, tmp );
		// Next six lines are to put X + Y + Z into length
		shuff = _mm_shuffle_ps( tmp, tmp, _MM_SHUFFLE( 1, 2, 3, 0 ) );
		length = shuff;
		shuff = _mm_shuffle_ps( tmp, tmp, _MM_SHUFFLE( 2, 3, 0, 1 ) );
		length = _mm_add_ss( length, shuff );
		shuff = _mm_shuffle_ps( tmp, tmp, _MM_SHUFFLE( 3, 0, 1, 2 ) );
		length = _mm_add_ss( length, shuff );
		// Take the reciprocal square root
		length = _mm_rsqrt_ss( length );
		// Move the rsqrt into all four positions
		length = _mm_shuffle_ps( length, length, _MM_SHUFFLE( 0, 0, 0, 0 ) );
		// Multiply it by data
		a = _mm_mul_ps( a.Data(), length );
	}

	Point3D Origin()
	{
		return Point3D( _mm_setzero_ps() );
	}

	Point3D PlaneNormal( const Point3D& p1, const Point3D& p2, const Point3D& p3 )
	{
		register __m128 a, b, tmp1, tmp2, shuff1, shuff2;
		
		a = _mm_sub_ps( p1.Data(), p2.Data() );
		b = _mm_sub_ps( p3.Data(), p1.Data() );
		// Shuffles a into Y,Z,X format
		shuff1 = _mm_shuffle_ps( a, a, _MM_SHUFFLE( 3, 0, 2, 1 ) );
		// Shuffles b into Z,X,Y format
		shuff2 = _mm_shuffle_ps( b, b, _MM_SHUFFLE( 1, 1, 0, 2 ) );
		tmp1 = _mm_mul_ps( shuff1, shuff2 );
		shuff1 = _mm_shuffle_ps( a, a, _MM_SHUFFLE( 1, 1, 0, 2 ) );
		shuff2 = _mm_shuffle_ps( b, b, _MM_SHUFFLE( 3, 0, 2, 1 ) );
		tmp2 = _mm_mul_ps( shuff1, shuff2 );
		a = _mm_sub_ps( tmp1, tmp2 );
		
		// And normalize 'a':
		b = _mm_mul_ps( a, a );
		
		b = _mm_add_ps( b, _mm_shuffle_ps( b, b, _MM_SHUFFLE( 2, 1, 0, 3 ) ) );
		b = _mm_add_ps( b, _mm_shuffle_ps( b, b, _MM_SHUFFLE( 2, 2, 2, 2 ) ) );
		
		b = _mm_rsqrt_ss( b );
		b = _mm_shuffle_ps( b, b, _MM_SHUFFLE( 0, 0, 0, 0 ) );

		return _mm_mul_ps( a, b );
	}
	
	Point3D PlaneUnormal( const Point3D& p1, const Point3D& p2, const Point3D& p3 )
	{
		register __m128 a, b, tmp1, tmp2, shuff1, shuff2;
	
		a = _mm_sub_ps( p1.Data(), p2.Data() );
		b = _mm_sub_ps( p3.Data(), p1.Data() );
		// Shuffles a into Y,Z,X format
		shuff1 = _mm_shuffle_ps( a, a, _MM_SHUFFLE( 3, 0, 2, 1 ) );
		// Shuffles b into Z,X,Y format
		shuff2 = _mm_shuffle_ps( b, b, _MM_SHUFFLE( 1, 1, 0, 2 ) );
		tmp1 = _mm_mul_ps( shuff1, shuff2 );
		shuff1 = _mm_shuffle_ps( a, a, _MM_SHUFFLE( 1, 1, 0, 2 ) );
		shuff2 = _mm_shuffle_ps( b, b, _MM_SHUFFLE( 3, 0, 2, 1 ) );
		tmp2 = _mm_mul_ps( shuff1, shuff2 );
		return _mm_sub_ps( tmp1, tmp2 );
	}
}

#endif
