/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Create.hh"
#include "Mesh.hh"

#include <iostream>

using SharpConstruct::Create::File;
using SharpConstruct::Mesh;

File::LoadLWO::LoadLWO( Mesh& mesh, const std::string& filename )
: mesh_( mesh ), file_( filename.c_str(), std::ios::in | std::ios::binary )
{}
				
double File::LoadLWO::Progress() const
{
	return 1;
}

bool File::LoadLWO::Step()
{
	// FIXME
	const bool FAIL( false );

	// Check if this is really an LWO
	if( !compare_chunk_( read_u4_(), "FORM" ) )
	{
		std::cerr << "FORM chunk not found" << std::endl;
		return FAIL;
	}
	read_u4_();
	if( !compare_chunk_( read_u4_(), "LWO2" ) )
	{
		std::cerr << "LWO2 chunk not found" << std::endl;
		return FAIL;
	}

	// Now iterate through remaining chunks
	while( file_.good() )
	{
		const unsigned cur( read_u4_() );
		const unsigned size( read_u4_() );

		if( compare_chunk_( cur, "PNTS" ) )
		{
			mesh_.VertexLocations().reserve( size / 12 );
			for( unsigned vi = 0; vi < size / 12; ++vi )
			{
				const float c1( read_float_() ),
					    c2( read_float_() ),
					    c3( read_float_() );
				mesh_.VertexLocations().push_back
					( Point3D( c1, c3, c2 ) );
			}
		}
		else if( compare_chunk_( cur, "POLS" ) )
		{
			offset_ = 0;

			// Ignore non-face polys
			const unsigned type( read_u4_() );
			if( compare_chunk_( type, "FACE" ) )
			{
				while( offset_ < size )
				{
					int cnt( read_u2_() );
					cnt &= 0x03FF;

					if( cnt == 3 )
						mesh_.Triangles().push_back( Triangle
							( read_sx_(), read_sx_(), read_sx_() ).ReverseVertices() );
					else if( cnt == 4 )
						mesh_.Quads().push_back( Quad
							( read_sx_(), read_sx_(), read_sx_(), read_sx_() ).ReverseVertices() );
					else
					{
						// FIXME
						std::vector< unsigned > polys;
						for( int i = 0; i < cnt; ++i )
							polys.push_back( read_u2_() );
						for( int poly = 2; poly < cnt; ++poly )
						{
							mesh_.Triangles().push_back( Triangle
								( polys[ 0 ], polys[ poly - 1 ], polys[ poly ] ) );
						}
					}
				}
			}
		}
		else
		{
			// In all other cases, just skip it
			file_.ignore( size );
		}
	}

	return false;
}

std::vector< char > File::LoadLWO::read_chars_( const unsigned n )
{
	std::vector< char > data;
	for( unsigned i = 0; i < n; ++i )
		data.push_back( file_.get() );

	// Reverse data on little-endian machines
	if( G_BYTE_ORDER == G_LITTLE_ENDIAN )
		reverse( data.begin(), data.end() );

	return data;
}

unsigned short File::LoadLWO::read_u2_()
{
	const Data data( read_chars_( 2 ) );
	unsigned short t;
	for( unsigned i = 0; i < 2; ++i )
		( ( char* )( &t ) )[ i ] = data[ i ];
	offset_ += 2;
	return t;
}
unsigned int File::LoadLWO::read_u4_()
{
	const Data data( read_chars_( 4 ) );
	unsigned int t;
	for( unsigned i = 0; i < 4; ++i )
		( ( char* )( &t ) )[ i ] = data[ i ];
	offset_ += 4;
	return t;
}
int File::LoadLWO::read_sx_()
{
	const unsigned short t1( read_u2_() );
	if( t1 < 0xFF00 )
		return t1;
	else
	{
		file_.unget();
		offset_ += 2;
		return ( file_.get() << 16 ) |
		       ( file_.get() << 8 ) |
		         file_.get();
	}
}
float File::LoadLWO::read_float_()
{
	const Data data( read_chars_( 4 ) );
	float t;
	for( unsigned i = 0; i < 4; ++i )
		( ( char* )( &t ) )[ i ] = data[ i ];
	offset_ += 4;
	return t;
}
std::string File::LoadLWO::read_string_( const unsigned n )
{
	const Data data( read_chars_( n ) );
	std::string t;
	for( unsigned i = 0; i < n; ++i )
		t.push_back( data[ i ] );
	return t;
}

bool File::LoadLWO::compare_chunk_( const unsigned ch1, const std::string& ch2 )
{
	return ( ( ch2[ 0 ] << 24 ) |
		 ( ch2[ 1 ] << 16 ) |
		 ( ch2[ 2 ] <<  8 ) |
		 ( ch2[ 3 ] ) ) == static_cast< signed >( ch1 );
}
